@@ -5,7 +5,7 @@ __info__ = info_manager(
     project = "PyCAMIA",
     package = "batorch",
     fileinfo = "Tensor functions for `batorch`.",
-    requires = "torch"
+    requires = ["torch", "pycamia", "pyoverload", "matplotlib"]
 )
 
 __all__ = """
@@ -34,6 +34,7 @@ __all__ = """
     cross_matrix
     uncross_matrix
     summary
+    display
 """.split()
 
 import math
@@ -43,7 +44,8 @@ with __info__:
     from pyoverload import *
     from pycamia import restore_type_wrapper, avouch, alias
     from pycamia import get_environ_vars, get_snakenames, get_args_expression
-    from pycamia import SPrint, Version
+    from pycamia import SPrint, Version, argmin, tokenize
+    from matplotlib import pyplot as plt
     
 def batorch_tensor(x):
     if not isinstance(x, bt.torch.Tensor): return bt.tensor(x)
@@ -431,48 +433,105 @@ def uncross_matrix(cross_matrix):
     return axis
 
 class summary:
-    def __init__(self, x, thumb = False):
-        var_name = get_args_expression("summary").split(',')[0].strip().split(')')[0]
-        out = SPrint(f"Summary of {var_name}:\n" + "=" * 50 + '\n' + x.__class__.__name__)
-        g = None
-        if not isinstance(x, bt.torch.Tensor): x = batorch_tensor(x)
-        elif not isinstance(x, bt.Tensor):
-            if x.is_leaf and x.grad is not None: g = batorch_tensor(x.grad.clone().detach())
-            x = batorch_tensor(x)
-        xshape = str(x.shape).split('Size')[-1]
-        out(f"{xshape}: [{x.n_special}] + {x.n_space} D tensor of values between [{round(x.min().item(), 4)}, {round(x.max().item(), 4)}].")
-        out(f"Device: {x.device}; Type: {x.type()}; Using gradient: {x.requires_grad}; Gradient func: {x.grad_fn}.")
-        if g is None and (not x.is_leaf or x.grad is None): grad_smy = None
-        else:
-            if g is None: g = x.grad.clone().detach()
-            gshape = str(g.shape).split('Size')[-1]
-            grad_smy = f"Gradient: {gshape} tensor of values [{round(g.min().item(), 4)}, {round(g.max().item(), 4)}].\n"
-            if g.n_space >= 2:
-                if g.has_batch: g = g.pick(0, [])
-                if g.has_channel: g = g.pick(0, {})
-            if g.n_dim == 1: a = min(4, g.size(0)); core_gradient = bt.crop_as(g, (a,)); label = "%ds"%a
-            elif g.n_dim == 2: a = min(4, g.size(0)); b = min(4, g.size(1)); core_gradient = bt.crop_as(g, (a, b)); label = "%d x %d" % (a, b)
-            else: a = min(4, g.size(-2)); b = min(4, g.size(-1)); core_gradient = bt.crop_as(g.flatten(0, g.n_dim - 3)[0], (a, b)); label = "%d x %d" % (a, b)
-            grad_smy += f"Core {label}-window for gradient: \n{core_gradient}\n".replace('Tensor', ' '*6).split(']'*(len(label)//2))[0] + ']'*(len(label)//2)
-        values = x.unique().tolist()
-        if x.nele >> 24 <= 0: out("Quantiles:", str(x.float().quantile(bt.arange(11).float()/10)).replace('Tensor(', '').split(']')[0] + ']')
-        else: out("Quantiles (of first 2^24 elements):", str(x.flatten(0, x.ndim-1).float()[:16777216].quantile(bt.arange(11).float()/10)).replace('Tensor(', '').split(']')[0] + ']')
-        if len(values) <= 10: out(f"Range ({len(values)} unique element{'s' if len(values) > 1 else ''}): [{', '.join(str(round(x, 4)) for x in values)}].")
-        else: out(f"Range ({len(values)} unique elements): \n[{', '.join(str(round(x, 4)) for x in values[:5])}, ..., {', '.join(str(round(x, 4)) for x in values[-5:])}].")
-        if x.n_space >= 2:
-            if x.has_batch: x = x.pick(0, [])
-            if x.has_channel: x = x.pick(0, {})
-        label = "4 x 4"
-        if x.n_dim == 1: a = min(4, x.size(0)); core = bt.crop_as(x, (a,)); label = "%ds"%a
-        elif x.n_dim == 2: a = min(4, x.size(0)); b = min(4, x.size(1)); core = bt.crop_as(x, (a, b)); label = "%d x %d" % (a, b)
-        else: a = min(4, x.size(-2)); b = min(4, x.size(-1)); core = bt.crop_as(x.flatten(0, x.n_dim - 3)[0], (a, b)); label = "%d x %d" % (a, b)
-        out(f"Core 4-window (the central {label}): \n{core}".replace('Tensor', ' '*6).split(']'*(len(label)//2))[0] + ']'*(len(label)//2))
-        if thumb:
-            gaps = bt.floor((bt.channel_tensor(list(x.space)) - 1) / 3).clamp(1).long()
-            down_sample = x[(image_grid(*(4 if s > 4 else s for s in x.space)) * gaps).split(1, {}, squeeze=True)]
-            out(f"Thumb 4-window (down sample to 4{' x 4'*(x.n_space-1)}): \n{down_sample}".replace('Tensor', ' '*6).split(']'*x.n_space)[0] + ']'*x.n_space)
-        if grad_smy is not None: out(grad_smy)
-        self.smy = out.text
-    def show(self): print(self.smy)
+    def __init__(self, *Xs, show_thumb = False, show_core = False, plot = False):
+        self.smys = []
+        self.show_as_plot = plot
+        if plot:
+            n_plots = len(Xs)
+            self.subplot_rows = math.floor(math.sqrt(n_plots))
+            self.subplot_cols = math.ceil(n_plots / self.subplot_rows)
+            self.canvases = []
+        if not hasattr(self, 'var_names'):
+            self.var_names = [vn.strip() for vn in tokenize(get_args_expression("summary"), sep=', ') if '=' not in vn]
+        for i, x in enumerate(Xs):
+            var_name = self.var_names[i]
+            class_name = x.__class__.__name__
+            g = None
+            if not isinstance(x, bt.torch.Tensor): x = batorch_tensor(x)
+            elif not isinstance(x, bt.Tensor):
+                if x.is_leaf and x.grad is not None: g = batorch_tensor(x.grad.clone().detach())
+                x = batorch_tensor(x)
+            batch_dim = f"[{x.n_special}]+" if x.n_special > 0 else ''
+            xshape = str(x.shape).split('Size')[-1]
+            out = SPrint(f">>Summary of {var_name}<<\n")
+            out(f"【Size】{batch_dim}{x.n_space}D {class_name} of size {xshape}")
+            out(f"【Device】{x.device}")
+            out(f"【Type】{x.type()}")
+            out(f"【Using gradient】{x.requires_grad}")
+            if x.requires_grad: out(f"【Gradient func】{x.grad_fn}")
+            nan_mask = bt.isnan(x)
+            num_nans = nan_mask.sum().sum()
+            if num_nans > 0:
+                out(f"【Number of NaNs】{num_nans.long().item()}")
+                out(f"【Center of NaNs】{bt.Size([round(x, 1) for x in (bt.image_grid(*x.shape) * nan_mask.remove_special_()).float().mean().tolist()]).special_from_(x)}")
+                grad_smy = None
+            else:
+                out(f"【MinMax】[{round(x.min().item(), 4)}, {round(x.max().item(), 4)}];")
+                if g is None and (not x.is_leaf or x.grad is None): grad_smy = None
+                else:
+                    if g is None: g = x.grad.clone().detach()
+                    gshape = str(g.shape).split('Size')[-1]
+                    grad_smy = f"【Gradient】{gshape} tensor of values [{round(g.min().item(), 4)}, {round(g.max().item(), 4)}]\n"
+                    if g.n_space >= 2:
+                        if g.has_batch: g = g.pick(0, [])
+                        if g.has_channel: g = g.pick(0, {})
+                    if g.n_dim == 1: a = min(4, g.size(0)); core_gradient = bt.crop_as(g, (a,)); label = "%ds"%a
+                    elif g.n_dim == 2: a = min(4, g.size(0)); b = min(4, g.size(1)); core_gradient = bt.crop_as(g, (a, b)); label = "%d x %d" % (a, b)
+                    else: a = min(4, g.size(-2)); b = min(4, g.size(-1)); core_gradient = bt.crop_as(g.flatten(0, g.n_dim - 3)[0], (a, b)); label = "%d x %d" % (a, b)
+                    if show_core: grad_smy += f"【Core {label}-window for gradient】\n{core_gradient}\n".replace('Tensor', ' '*6).split(']'*(len(label)//2))[0] + ']'*(len(label)//2)
+                values = x.unique().tolist()
+                if x.nele >> 24 <= 0: out("【Quintiles】", [round(u.item(), 4) for u in x.float().quantile(bt.arange(6).float()/5)])
+                else: out("【Quintiles (of first 2^24 elements)】", str(x.flatten(0, x.ndim-1).float()[:16777216].quantile(bt.arange(6).float()/5)).replace('Tensor(', '').split(']')[0] + ']')
+                if len(values) <= 10:
+                    out(f"【Range ({len(values)} unique element{'s' if len(values) > 1 else ''})】")
+                    if len(values) > 6:
+                        out(f">> [{', '.join(str(round(x, 4)) for x in values)}]")
+                    else:
+                        num_pairs = [(round(v, 4), (x == v).sum().sum().item()) for v in values]
+                        out(f">> [{', '.join(f'{v}({n})' for v, n in num_pairs)}]")
+                else: out(f"【Range ({len(values)} unique elements)】\n  [{', '.join(str(round(x, 4)) for x in values[:5])}, ..., {', '.join(str(round(x, 4)) for x in values[-5:])}]")
+            if x.n_space >= 2:
+                if x.has_batch: x = x.pick(0, [])
+                if x.has_channel: x = x.pick(0, {})
+            label = "4 x 4"
+            if x.n_dim == 1: a = min(4, x.size(0)); core = bt.crop_as(x, (a,)); label = "%ds"%a
+            elif x.n_dim == 2: a = min(4, x.size(0)); b = min(4, x.size(1)); core = bt.crop_as(x, (a, b)); label = "%d x %d" % (a, b)
+            else: a = min(4, x.size(-2)); b = min(4, x.size(-1)); core = bt.crop_as(x.flatten(0, x.n_dim - 3)[0], (a, b)); label = "%d x %d" % (a, b)
+            if show_core: out(f"【Core 4-window (the central {label})】\n{core}".replace('Tensor', ' '*6).split(']'*(len(label)//2))[0] + ']'*(len(label)//2))
+            if show_thumb:
+                gaps = bt.floor((bt.channel_tensor(list(x.space)) - 1) / 3).clamp(1).long()
+                down_sample = x[(image_grid(*(4 if s > 4 else s for s in x.space)) * gaps).split(1, {}, squeeze=True)]
+                out(f"【Thumb 4-window (down sampled to 4{' x 4'*(x.n_space-1)})】\n{down_sample}".replace('Tensor', ' '*6).split(']'*x.n_space)[0] + ']'*x.n_space)
+            if grad_smy is not None: out(grad_smy)
+            self.smys.append(out.text)
+            if plot:
+                while x.n_dim > 2:
+                    x = x.pick(0, argmin(list(x.shape))[0])
+                self.canvases.append(x.detach())
+    def show(self):
+        if not self.show_as_plot:
+            print(self)
+            return
+        plt.figure(figsize=(15, 7))
+        for i, x in enumerate(self.canvases):
+            ax = plt.subplot(self.subplot_rows, self.subplot_cols, i+1)
+            old_size = plt.rcParams['font.size']
+            plt.rcParams['font.size'] = 8
+            if x.n_dim == 2:
+                img = plt.imshow(x, cmap=plt.cm.gray)
+                plt.colorbar(mappable=img, ax=ax, orientation='vertical')
+                plt.text(-0.5, -0.5, self.smys[i])
+            elif x.n_dim == 1:
+                plt.plot(bt.arange(len(x)), x)
+                plt.text(0, plt.ylim()[1], self.smys[i])
+            else: raise TypeError("Failed to display 0-dimensional data in plot.")
+            plt.rcParams['font.size'] = old_size
+        plt.show()
     @alias("__repr__")
-    def __str__(self): return self.smy
+    def __str__(self): return ''.join(self.smys)
+
+class display(summary):
+    def __init__(self, *args, **kwargs):
+        kwargs['plot'] = True
+        self.var_names = [vn.strip() for vn in tokenize(get_args_expression("display"), sep=', ') if '=' not in vn]
+        return super().__init__(*args, **kwargs)
